#ifndef MSP_CORE_HASH_H_
#define MSP_CORE_HASH_H_

#include <cstdint>
#include <limits>
#include <string>

namespace Msp {

template<unsigned N, unsigned L = 3+(N>8)+(N>16)+(N>32)+(N>64)>
struct HashTraits;

template<>
struct HashTraits<64, 6>
{
	using HashType = uint64_t;
	using ResultType = uint64_t;
	static constexpr HashType prime = 1099511628211ULL;
	static constexpr HashType offset = 14695981039346656037ULL;
	static constexpr bool folded = false;
};

template<>
struct HashTraits<32, 5>
{
	using HashType = uint32_t;
	using ResultType = uint32_t;
	static constexpr HashType prime = 16777619;
	static constexpr HashType offset = 2166136261U;
	static constexpr bool folded = false;
};

template<unsigned N>
struct HashTraits<N, 6>
{
	using HashType = uint64_t;
	using ResultType = uint64_t;
	static constexpr bool folded = true;
};

template<unsigned N>
struct HashTraits<N, 5>
{
	using HashType = uint32_t;
	using ResultType = uint32_t;
	static constexpr bool folded = true;
};

template<unsigned N>
struct HashTraits<N, 4>
{
	using HashType = uint32_t;
	using ResultType = uint16_t;
	static constexpr bool folded = true;
};

template<unsigned N>
struct HashTraits<N, 3>
{
	using HashType = uint32_t;
	using ResultType = uint8_t;
	static constexpr bool folded = true;
};

/**
Computes a single round of a hash, adding one byte of data.  N must be a native
hash size.
*/
template<unsigned N>
typename HashTraits<N>::ResultType hash_round(typename HashTraits<N>::HashType result, uint8_t byte)
{
	constexpr typename HashTraits<N>::HashType prime = HashTraits<N>::prime;
	return (result^byte)*prime;
}

/**
Updates a previously computed hash by adding bytes to it.  N must be a native
hash size.
*/
template<unsigned N>
typename HashTraits<N>::ResultType hash_update(typename HashTraits<N>::HashType result, const void *data, std::size_t size)
{
	const uint8_t *ptr = static_cast<const uint8_t *>(data);
	for(unsigned i=0; i<size; ++i)
		result = hash_round<N>(result, *ptr++);
	return result;
}

template<unsigned N>
typename HashTraits<N>::ResultType hash_update(typename HashTraits<N>::HashType result, const std::string &str)
{ return hash_update<N>(result, str.data(), str.size()); }

// For some reason this causes an ambiguity error without the enable_if.
template<unsigned N, typename T>
typename std::enable_if<!std::is_same<T, std::string>::value, typename HashTraits<N>::ResultType>::type hash_update(typename HashTraits<N>::HashType result, const T &obj)
{ return hash_update<N>(result, &obj, sizeof(obj)); }

/**
Manually folds a hash to a smaller size.
*/
template<unsigned N, typename T>
typename std::enable_if<N!=std::numeric_limits<T>::digits, typename HashTraits<N>::ResultType>::type hash_fold(T hash)
{ return (hash>>N ^ hash) & ((T(1)<<N)-1); }

template<unsigned N, typename T>
typename std::enable_if<N==std::numeric_limits<T>::digits, typename HashTraits<N>::ResultType>::type hash_fold(T hash)
{ return hash; }

/**
Computes an N-bit Fowler-Noll-Vo (FNV-1a) hash.  If N is not a native hash
size, the next larger native hash is computed and XOR-folding is applied to
the result.

http://en.wikipedia.org/wiki/Fowler-Noll-Vo_hash_function
http://www.isthe.com/chongo/tech/comp/fnv/index.html
*/
template<unsigned N>
typename HashTraits<N>::ResultType hash(const void *data, std::size_t size)
{
	constexpr unsigned hash_bits = std::numeric_limits<typename HashTraits<N>::HashType>::digits;
	typename HashTraits<N>::HashType result = hash_update<hash_bits>(HashTraits<hash_bits>::offset, data, size);
	if(HashTraits<N>::folded)
		result = hash_fold<N>(result);
	return result;
}

/**
Convenience function to compute an N-bit hash of a string.
*/
template<unsigned N>
typename HashTraits<N>::ResultType hash(const std::string &str)
{ return hash<N>(str.data(), str.size()); }

/**
Convenience function to compute an N-bit hash of any object.  The entire object
representation is hashed, including any padding bytes.
*/
template<unsigned N, typename T>
typename HashTraits<N>::ResultType hash(const T &obj)
{ return hash<N>(&obj, sizeof(obj)); }

} // namespace Msp

#endif
