#ifndef MSP_IO_ZLIBCOMPRESSED_H_
#define MSP_IO_ZLIBCOMPRESSED_H_

#include <stdexcept>
#include <string>
#include <sigc++/trackable.h>
#include "base.h"

namespace Msp {
namespace IO {

class zlib_error: public std::runtime_error
{
private:
	int m_code;

public:
	zlib_error(const std::string &, int);
	~zlib_error() throw() override = default;

	int code() const throw() { return m_code; }
};

/**
Compresses or decompresses data with zlib.  This class is a filter that
operates on top of another I/O object.

To ensure proper termination of the compressed data stream, the ZlibCompressed
object must be destroyed before the underlying object is closed.
*/
class ZlibCompressed: public Base, public sigc::trackable
{
private:
	struct Private;

	Base &below;
	std::size_t buffer_size = 1024;
	unsigned char *in_buffer = nullptr;
	unsigned char *out_buffer = nullptr;
	bool stream_end = false;
	Private *priv = nullptr;

public:
	/** Creates a zlib de/compression object.  The underlying object must be
	open for reading or writing, not both.  The level parameter determines
	compression quality, ranging from 1 (fastest) to 9 (best compression). */
	ZlibCompressed(Base &b, unsigned level = 9): ZlibCompressed(b, b.get_mode()&M_RDWR, level) { }

	/** Creates a zlib de/compression object.  Mode must be either read or
	write, and compatible with the underlying object. */
	ZlibCompressed(Base &, Mode, unsigned level = 9);

	~ZlibCompressed() override;

	void set_block(bool) override;
	void set_inherit(bool) override;

	void flush();

protected:
	std::size_t do_write(const char *, std::size_t) override;

private:
	/** Compresses data and writes it to the underlying object.  Returns true if
	progress was made, false otherwise.  flush_mode can be any of zlib's flush
	modes.  If it is Z_FINISH, false is also returned when the stream has been
	terminated. */
	bool compress_data(int flush_mode);

public:
	std::size_t do_read(char *, std::size_t) override;

	const Handle &get_handle(Mode) override;
};

} // namespace IO
} // namespace Msp

#endif
